/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * File: mock_backend.cpp
 *
 * A mock backend for P4, used for early development of plugins and the network
 * client. Only supports a single connection between two nodes - one should be
 * the server, and one should be the client. For more information, see
 * p4mockclient.h.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */


#include "p4client.h"
#include "p4error.h"
#include "debug_print.h"

#include <stdio.h>      // for perror
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>  // for inet_ntoa, inet_aton
#include <netdb.h>      // for gethostbyname
#include <unistd.h>     // for close
#include <errno.h>
#include <assert.h>

#include <iostream>
using namespace std;

#ifndef INADDR_NONE
#define INADDR_NONE static_cast<ulong>(-1)  // for a few platforms, eg. Solaris
#endif

/*****************************************************************************
 * CONSTANTS
 *****************************************************************************/
const uint kmsg_buf_size = 256;
const char *kp4_preamble = "P4 CONNECT/"VERSION;
const char *kp4_preamble_ack = "P4 OK";
const ushort plugin_id = 2481;


/*****************************************************************************
 * GLOBAL VARS
 *****************************************************************************/
static int gsock = -1;
static char *gaddress = NULL;
static ushort gport = 0;


/*****************************************************************************
 * PRIVATE HELPER PROTOTYPES
 *****************************************************************************/
void error(char *s);
uint mock_listen();
struct in_addr get_local_addr();
struct in_addr resolve_addr(char *addr);



void p4_mock_setup(char *address, ushort port)
{
  gport = port;
  if (address)
    gaddress = strdup(address);
  else
    gaddress = NULL;
}


/*****************************************************************************
 * PUBLIC FUNCTIONS FROM THE CLIENT API
 *****************************************************************************/

uint p4_bootstrap()
{
  debug_print("NOTE: using the mock P4 backend, libmockp4!\n");

  if (gsock >= 0)
    error("already connected, cannot connect again");
  if (gport == 0)
    error("you must call p4_mock_setup() before p4_bootstrap()");

  // do we have an address to connect to? if not, listen for connections
  if (!gaddress)
    return mock_listen();


  // fill in local address
  struct sockaddr_in local, remote;
  local.sin_addr = get_local_addr();
  local.sin_family = AF_INET;
  local.sin_port = htons(gport);


  // fill in remote address
  remote.sin_addr = resolve_addr(gaddress);
  if (remote.sin_addr.s_addr == INADDR_NONE)
    error("invalid address");
  remote.sin_family = AF_INET;
  remote.sin_port = htons(gport);


  // create socket and connect
  if ((gsock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    error("could not create socket");

  if (connect(gsock, (struct sockaddr *)&remote, sizeof(remote)) < 0) {
    debug_print("couldn't connect to %s:%d\n", gaddress, gport);
    p4_quit();
    return mock_listen();
  }

  // do preamble
  char msg[kmsg_buf_size];
  int ret;

  send(gsock, kp4_preamble, strlen(kp4_preamble) + 1, 0);
  ret = recv(gsock, msg, kmsg_buf_size - 1, 0);
  if (ret <= 0)
    return ret;      // error or other side closed connection

  msg[strlen(kp4_preamble_ack)] = '\0';
  if (strcmp(msg, kp4_preamble_ack)) {
    debug_print("client did not complete preamble, dropping connection.\n");
    return kerr_connect_failed;
  }

  return kerr_success;
}


uint mock_listen()
{
  struct sockaddr_in local, remote;
  uint remote_size = sizeof(remote);


  if (gsock >= 0)
    error("already connected, cannot connect again");


  while (true) {          // listen until we get the preamble right
    // fill in local address
    local.sin_addr = get_local_addr();
    local.sin_family = AF_INET;
    local.sin_port = htons(gport);


    // create and bind socket
    if ((gsock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
      error("could not create socket");

    if (bind(gsock, (struct sockaddr *)&local, sizeof(local)) < 0)
      error("could not bind socket");


    // listen for and accept incoming connections
    if (listen(gsock, 1) < 0)
      error("couldn't listen");

    debug_print("listening for connections on port %d\n", gport);

    // accept an incoming connection
    int newsock;

    newsock = accept(gsock, (struct sockaddr *)&remote, &remote_size);
    close(gsock);
    gsock = newsock;

    if (gsock < 0) {
      debug_print("couldn't accept incoming connection");
      continue;
    }

    debug_print("incoming connection from %s\n", inet_ntoa(remote.sin_addr));


    // attempt P4 preamble
    char msg[kmsg_buf_size];
    int ret;

    ret = recv(gsock, msg, 256, 0);
    
    if (ret <= 0) {
      debug_print("recv error");
      p4_quit();
      return ret;
    }

    msg[strlen(kp4_preamble)] = '\0';
    if (strcmp(msg, kp4_preamble)) {    // not preamble string!
      debug_print("client did not complete preamble, dropping connection.\n");
      p4_quit();
      return kerr_couldnt_connect;
    }

    // got preamble, send response
    send(gsock, kp4_preamble_ack, strlen(kp4_preamble_ack) + 1, 0);
    break;
  }

  // won't reach here
  return kerr_success;
}



/*****************************************************************************
 * PUBLIC FUNCTIONS - FROM NORMAL PLUGIN API
 *****************************************************************************/
uint p4_broadcast(ushort plugin_id, const char *data, int size, bool encrypted)
{
  return p4_send(plugin_id, NULL, data, size, encrypted);
}


uint p4_quit()
{
  if (gsock >= 0)
    close(gsock);
  gsock = -1;

  return kerr_success;
}


uint p4_send(ushort plugin_id, const char *user, const char *data, int size,
             bool encrypted)
{
  int sent = 0;
  int ret;

  assert(data && size > 0);

  // send until the whole thing goes through
  while (sent < size) {
    ret = send(gsock, data + sent * sizeof(char), size - sent, 0);
    if (ret == -1)
      return 1;
    sent += ret;
  }

  return 0;
}


uint p4_get_num_plugins(const char *user, int *num_plugins)
{
  assert(user && num_plugins);

  *num_plugins = (gsock == -1) ? 0 : 1;
  return kerr_success;
}


uint p4_get_plugins(const char *user, ushort *plugins, ushort num_plugins,
                    ushort *num_written)
{
  assert(plugins && num_written && num_plugins > 0);

  if (gsock == -1) {      // not connected
    *num_written = 0;
    return kerr_lka_user_not_found;
  }

  // only one plugin
  *plugins = plugin_id;
  *num_written = 1;
  return kerr_success;
}

uint p4_get_latency(const char *user, int *latency)
{
  assert(latency);

  if (gsock == -1)
    return kerr_lka_user_not_found;

  *latency = 100;
  return kerr_success;
}

uint p4_get_num_users(ushort plugin_id, int *num_users)
{
  assert(plugin_id  == (plugin_id));

  *num_users = (gsock == -1) ? 0 : 1;
  return kerr_success;
}

uint p4_get_users(ushort plugin_id, char **users, int num_users,
                  int *num_written)
{
  assert(plugin_id  == (plugin_id) && num_written &&
         users && users[0] && num_users > 0);

  if (gsock == -1) {
    *num_written = 0;
    return kerr_success;
  }

  sprintf(users[0], "yourmom%d", rand() % 10);
  *num_written = 1;
  return 1;
}

uint p4_authenticate(ushort plugin_id, bool *authenticated)
{
  debug_print("p4_authenticate is not implemented\n");
  return 1;
}


uint p4_peek(ushort plugin_id, int *size)
{
  fd_set sockfd;
  struct timeval timeout = { 0, 0 };

  assert(size);

  // select on gsock
  FD_ZERO(&sockfd);
  FD_SET(gsock, &sockfd);
  select(gsock + 1, &sockfd, NULL, NULL, &timeout);

  // is there data ready?
  *size = (FD_ISSET(gsock, &sockfd)) ? 99999 : 0;
  return 1;
}


/**
 * note that this call blocks until data is received!
 */
uint p4_receive(ushort plugin_id, char *user, char *data, int size,
                int *size_written, bool *encrypted)
{
  int ret;

  assert(data && size > 0 && size_written && encrypted);

  // read from socket
  ret = recv(gsock, data, size, 0);
  if (ret <= 0)
    error("recv error, exiting");
  else
    *size_written = ret;

  return 0;
}



/*****************************************************************************
 * PRIVATE FUNCTIONS
 *****************************************************************************/

/** error
 *
 * Prints out the given string and the current error message, then exits.
 */
void error(char *s)
{
  perror(s);
  exit(-1);
}


/** get_local_addr
 *
 * Returns the local IP address in in_addr form (in *host* byte order).
 */
struct in_addr get_local_addr()
{
  struct hostent *host;
  struct in_addr sin;

  // get local address
  host = gethostbyname("localhost");
  assert(host);

  //sa->sin_addr.s_addr = htonl(INADDR_ANY);
  assert(host->h_length == sizeof(sin));
  memcpy(&sin.s_addr, host->h_addr_list[0], sizeof(sin.s_addr));
  return sin;
}


/** resolve_addr
 *
 * Resolves a hostname (e.g. blaize.stanford.edu) or IP address (e.g.
 * 128.12.75.93) to an in_addr. If the given address is invalid, returns
 * INADDR_NONE.
 */
struct in_addr resolve_addr(char *addr)
{
  struct hostent *host;
  struct in_addr sin;

  if ((host = gethostbyname(addr))) {  // addr was a valid hostname
    memcpy(&sin.s_addr, host->h_addr_list[0], sizeof(sin.s_addr));
    return sin;
  } else if (inet_aton(addr, &sin)) {
    return sin;            // addr was a valid IP address
  } else {
    sin.s_addr = INADDR_NONE;
    return sin;
  }
}



